//+------------------------------------------------------------------+
//|                                                    Exception.mqh |
//|                                 Copyright  2011-2016, Marketeer |
//|                          https://www.mql5.com/en/users/marketeer |
//| MQL's OOP notes: rubber arrays, safe numbers,                    |
//|                     exceptions handling and other tricks: part 2 |
//|                        https://www.mql5.com/en/blogs/post/683243 |
//+------------------------------------------------------------------+

#ifndef EXCEPTIONS
#define EXCEPTIONS

class Exception
{
  public:
    const string location;
    const string message;
    double result;
    
    Exception(string l, string m): location(l), message(m), result(0) {}
    
    void throw(const double r)
    {
      result = r;
    }
};

class catch;

class try
{
  static try *landmarks[];
  static int cursor;
  
  string location;
  catch *ptrc;
  
  public:
    try(const string mark)
    {
      location = mark;
      int n = ArraySize(landmarks);
      ArrayResize(landmarks, n + 1);
      landmarks[n] = &this;
      Print("try at ", mark, " ", n);
    }
    
    int activate()
    {
      int back = cursor;
      int size = ArraySize(landmarks);
      for(int i = 0; i < size; i++)
      {
        if(landmarks[i] == &this)
        {
          cursor = i;
          return back;
        }
      }
      return back;
    }
    
    void restore(int rollback)
    {
      cursor = rollback;
    }
    
    void bind(catch &c)
    {
      ptrc = &c;
    }

    static double throw(const string message)
    {
      int size = ArraySize(landmarks);
      if(cursor >= 0 && cursor < size)
      {
        if(CheckPointer(landmarks[cursor].ptrc) != POINTER_INVALID)
        {
          Exception e(landmarks[cursor].location, message);
          landmarks[cursor].ptrc.handle(e);
          return e.result;
        }
      }
      return 0;
    }
};

try *try::landmarks[];
int try::cursor = -1;

class catch
{
  public:
    catch(try &_try)
    {
      _try.bind(this);
    }
    
    virtual void handle(Exception &e)
    {
      // dump the error
    }
};

class context
{
  private:
    int restore;
    try *ptr;
    
  public:
    context(try &_try)
    {
      ptr = &_try;
      restore = ptr.activate();
    }
    
    ~context()
    {
      ptr.restore(restore);
    }
};

#define TRY { static try _try(__FILE__ + ", " + (string)__LINE__ + ", " + __FUNCSIG__); context _context(_try);
#define CATCH(A) class catchx:public catch{public: catchx(try &_x):catch(_x){} virtual void handle(Exception &A)
#define END }; static catchx _catchx(_try); }

#endif